(function (global, factory) {
    typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('serializr'), require('mobx')) :
    typeof define === 'function' && define.amd ? define(['exports', 'serializr', 'mobx'], factory) :
    (factory((global['mobx-decorated-models'] = {}),global.serializr,global.mobx));
}(this, (function (exports,serializr,mobx) { 'use strict';

    function isSerializable(model, property) {
      return !!(!model.$nonSerializable || !model.$nonSerializable[property]);
    }
    function markNonserializable(model, property) {
      model.$nonSerializable = model.$nonSerializable || {};
      model.$nonSerializable[property] = true;
    }

    function hasInterceptors(model) {
      return Boolean(model.constructor.$mxdm && model.constructor.$mxdm.interceptors.length);
    }

    function getSchema(klass) {
      if (klass.$mxdm) {
        return klass.$mxdm;
      }

      Object.defineProperty(klass, '$mxdm', {
        enumerable: false,
        writable: false,
        configurable: true,
        value: Object.freeze({
          interceptors: [],
          properties: new Map()
        })
      });
      return klass.$mxdm;
    }
    function addLazyInitializer(target, fn) {
      const _getSchema = getSchema(target.constructor),
            interceptors = _getSchema.interceptors;

      interceptors.push(fn);
    }
    function attachInterceptors(model) {
      if (!hasInterceptors(model) || model.$mxdbDidAttachInterceptors) {
        return;
      }

      Object.defineProperty(model, '$mxdbDidAttachInterceptors', {
        enumerable: false,
        writable: false,
        configurable: true,
        value: true
      });
      getSchema(model.constructor).interceptors.forEach(fn => fn(model));
    }
    function setupModel({
      attrs,
      modelClass: Klass,
      array,
      defaultAttributes,
      inverseOf,
      parentModel,
      parentModelProp
    }) {
      if (!attrs) {
        return attrs;
      }

      const isObject = 'object' === typeof attrs;
      const canWrite = Object.isExtensible(attrs);

      if (isObject && canWrite) {
        if (defaultAttributes) {
          if ('function' === typeof defaultAttributes) {
            defaultAttributes = defaultAttributes.call(parentModel, array, parentModel);
          }

          Object.keys(defaultAttributes).forEach(key => {
            if (!attrs[key]) {
              attrs[key] = defaultAttributes[key];
            }
          });
        }

        if (inverseOf) {
          attrs[inverseOf] = parentModel;

          if (parentModelProp) {
            attrs[`${inverseOf}_association_name`] = parentModelProp;
          }
        }
      }

      const model = Klass && 'function' === typeof Klass && !(attrs instanceof Klass) ? new Klass(attrs) : attrs;

      if (inverseOf) {
        markNonserializable(model, inverseOf);
      }

      return model;
    }

    const ModelsMap = Object.create(null);

    const defaultModelLookup = ({
      property,
      model = property
    }) => {
      if ('function' === typeof model) {
        return model;
      }

      return ModelsMap[model] || ModelsMap[property];
    };

    let modelLookup = defaultModelLookup;

    const defaultRememberModel = model => {
      ModelsMap[model.identifiedBy || model.name] = model;
    };

    let rememberModel = defaultRememberModel;
    function registerModel(model) {
      rememberModel(model);
    }
    function findModel(model, property) {
      return modelLookup({
        model,
        property
      });
    }
    function rememberModelUsing(fn) {
      rememberModel = fn || defaultRememberModel;
    }
    function lookupModelUsing(fn) {
      modelLookup = fn || defaultModelLookup;
    }

    const PendingLookups = [];

    function objectSerializer() {
      return {
        serializer(obj, prop, parent) {
          return isSerializable(parent, prop) ? obj : serializr.SKIP;
        },

        deserializer(json, cb) {
          cb(null, json);
        }

      };
    }

    function addReference(parentModel, propName, options, cb) {
      const model = findModel(options.model, propName);

      if (model) {
        serializr.getDefaultModelSchema(parentModel).props[propName] = cb(model, propName, options);
      } else {
        serializr.getDefaultModelSchema(parentModel).props[propName] = objectSerializer();
        PendingLookups.push({
          parentModel,
          propName,
          options,
          cb
        });
      }
    }

    const dateSerializer = () => serializr.custom(date => date ? date.toJSON() : null, json => json ? new Date(json) : null);

    const CustomSerializerTypes = {
      date: dateSerializer,
      object: objectSerializer,
      array: serializr.list
    };

    function getSerializer(options, defaultSerializer) {
      if (options.serializer) {
        return serializr.custom(options.serializer[0], options.serializer[1]);
      }

      const fns = (CustomSerializerTypes[options.type] || defaultSerializer)();

      if (options.writeOnly) {
        return {
          serializer: () => serializr.SKIP,
          deserializer: fns.deserializer
        };
      }

      return fns;
    }

    const Fields = {
      identifier: options => getSerializer(options, serializr.identifier),
      field: options => getSerializer(options, serializr.primitive),
      session: options => getSerializer(Object.assign({}, options, {
        writeOnly: true
      }), serializr.primitive)
    };

    function associationSerializers(modelClass, propName, options) {
      return {
        serializer: options.serializer || modelClass.serialize || serializr.object(serializr.getDefaultModelSchema(modelClass)).serializer,
        deserializer: options.deserializer || modelClass.deserializer || ((attrs, context) => {
          let model; // double equals catches null/undefined

          if (null == attrs) {
            attrs = {};
          }

          if (attrs instanceof modelClass) {
            model = attrs;
          } else {
            model = setupModel(Object.assign({}, options, {
              attrs,
              modelClass
            }));
          }

          if (options.inverseOf) {
            markNonserializable(model, options.inverseOf);
            model[`${options.inverseOf}_association_name`] = propName;
            model[options.inverseOf] = context.target;
          }

          return model;
        })
      };
    }

    const Associations = {
      hasMany(modelClass, propName, options) {
        const _associationSerialize = associationSerializers(modelClass, propName, options),
              deserializer = _associationSerialize.deserializer,
              serializer = _associationSerialize.serializer;

        options.model = modelClass;
        return Object.assign({}, options, {
          deserializer(collection, cb, context) {
            const value = collection && collection.map ? collection.map(attrs => deserializer(attrs, context)) : collection;
            cb(null, value);
          },

          serializer(collection, name, parent) {
            if (!isSerializable(parent, name)) {
              return serializr.SKIP;
            }

            return collection && collection.map ? collection.map(model => serializer(model, name, parent)) : collection;
          }

        });
      },

      belongsTo(modelClass, propName, options) {
        const _associationSerialize2 = associationSerializers(modelClass, propName, options),
              deserializer = _associationSerialize2.deserializer,
              serializer = _associationSerialize2.serializer;

        options.model = modelClass;
        return Object.assign({}, options, {
          deserializer(attrs, cb, context) {
            cb(null, deserializer(attrs, context));
          },

          serializer(belongsTo, name, parent) {
            if (!isSerializable(parent, name)) {
              return serializr.SKIP;
            }

            return serializer(belongsTo, name, parent);
          }

        });
      }

    };
    const MixedInInstanceMethods = {
      serialize() {
        const schema = serializr.getDefaultModelSchema(this.constructor);
        return serializr.serialize(schema, this);
      },

      update(json, callback) {
        return serializr.update(serializr.getDefaultModelSchema(this.constructor), this, json, callback);
      }

    };
    const MixedInClassMethods = {
      deserialize(json, callback) {
        return serializr.deserialize(serializr.getDefaultModelSchema(this), json, callback);
      }

    };

    function decorateModel(model) {
      Object.assign(model.prototype, MixedInInstanceMethods);
      Object.assign(model, MixedInClassMethods);
      registerModel(model);
      const serializeSchema = {};

      const _getSchema = getSchema(model),
            properties = _getSchema.properties;

      properties.forEach(({
        type,
        options
      }, name) => {
        if (Fields[type]) {
          serializeSchema[name] = Fields[type](options, name);
        }
      }); // eslint-disable-next-line new-cap

      serializr.createModelSchema(model, serializeSchema, ({
        json
      }) => new model(json));
      properties.forEach(({
        type,
        options
      }, name) => {
        if (Associations[type] || options.model) {
          addReference(model, name, options, Associations[type] || Associations.belongsTo);
        }
      });

      for (let i = PendingLookups.length - 1; i >= 0; i -= 1) {
        const _PendingLookups$i = PendingLookups[i],
              parentModel = _PendingLookups$i.parentModel,
              propName = _PendingLookups$i.propName,
              options = _PendingLookups$i.options,
              cb = _PendingLookups$i.cb;
        const referencedModel = findModel(options.model, propName);

        if (referencedModel) {
          const parentModelSchema = serializr.getDefaultModelSchema(parentModel);
          parentModelSchema.props[propName] = cb(referencedModel, propName, options);
          PendingLookups.splice(i, 1);
        }
      }
    }

    function registerCustomType(id, serializer) {
      CustomSerializerTypes[id] = () => serializr.custom(serializer.serialize, serializer.deserialize);
    }
    function identifiedBy(modelId) {
      return model => {
        Object.defineProperty(model, 'identifiedBy', {
          value: modelId,
          writable: false
        });
        return decorateModel(model);
      };
    }
    function unresolvedAssociations() {
      return PendingLookups.map(({
        parentModel: model,
        propName: property
      }) => ({
        model,
        property
      }));
    }

    function onCollectionChangeInterceptor({
      modelClass,
      model,
      defaults: defaultAttributes,
      inverseOf
    }, parentModel, parentModelProp) {
      return change => {
        if (!change.newValue) {
          change.newValue = {};
        }

        if (!modelClass) {
          modelClass = findModel(model, parentModelProp);
        }

        if ('splice' === change.type) {
          for (let i = 0; i < change.added.length; i += 1) {
            change.added[i] = setupModel({
              attrs: change.added[i],
              array: change.object,
              modelClass,
              defaultAttributes,
              inverseOf,
              parentModel,
              parentModelProp
            });
          }
        } else if ('update' === change.type) {
          change.newValue = setupModel({
            attrs: change.newValue,
            array: change.object,
            modelClass,
            defaultAttributes,
            inverseOf,
            parentModel,
            parentModelProp
          });
        }

        return change;
      };
    }

    function createCollection(options, parentModel, parentModelProp) {
      const ary = mobx.observable.array([]);

      if (options) {
        if (options.model) {
          ary.intercept(onCollectionChangeInterceptor(options, parentModel, parentModelProp));
        }

        if (options.extend) {
          if ('function' === typeof options.extend) {
            options.extend(ary);
          } else {
            Object.keys(options.extend).forEach(k => {
              Object.defineProperty(ary, k, {
                value: options.extend[k]
              });
            });
          }
        }
      }

      return ary;
    }

    function onPropertySet(change, {
      modelClass,
      defaultAttributes,
      inverseOf,
      parentModel,
      parentModelProp
    }) {
      change.newValue = setupModel({
        attrs: change.newValue,
        modelClass,
        defaultAttributes,
        inverseOf,
        parentModel,
        parentModelProp
      });
      return change;
    }

    function onHasManySet(change, options) {
      if (change.type !== 'update' || !change.newValue) {
        return change;
      }

      const modelClass = options.modelClass,
            defaultAttributes = options.defaultAttributes,
            inverseOf = options.inverseOf,
            parentModel = options.parentModel,
            parentModelProp = options.parentModelProp;
      const array = mobx.observable.array([]);

      for (let i = 0; i < change.newValue.length; i += 1) {
        array.push(setupModel({
          attrs: change.newValue[i],
          array: change.newValue,
          modelClass,
          defaultAttributes,
          inverseOf,
          parentModel,
          parentModelProp
        }));
      }

      if (parentModel[parentModelProp]) {
        parentModel[parentModelProp].replace(array);
        return null;
      }

      change.newValue = mobx.observable.array(array);
      return change;
    }

    const Initializers = {
      object: () => mobx.observable.object({}),
      array: createCollection
    };
    const TypeInitializers = {
      hasMany: createCollection
    };

    function getInitializer(type, options, propName) {
      const fn = TypeInitializers[type] || Initializers[options.type];

      if (!fn) {
        return undefined;
      }

      return function () {
        // eslint-disable-line func-names
        return fn(options, this, propName);
      };
    }

    function installModelInterceptor(interceptingFn, {
      target,
      property
    }) {
      const _getSchema = getSchema(target.constructor),
            properties = _getSchema.properties;

      addLazyInitializer(target, model => {
        const schemaProps = properties.get(property).options;
        mobx.intercept(model, property, change => interceptingFn(change, {
          inverseOf: schemaProps.inverseOf,
          modelClass: schemaProps.model,
          parentModel: model,
          parentModelProp: property,
          defaultAttributes: schemaProps.defaults
        }));
      });
    }

    function addAttribute(type, setAttributeFn, target, property, descriptor, options = {}) {
      getSchema(target.constructor).properties.set(property, {
        name: property,
        type,
        options
      });
      const initializer = getInitializer(type, options, property);

      if (initializer) {
        descriptor.initializer = initializer;
      }

      const definition = mobx.observable(target, property, descriptor);

      if ('belongsTo' === type && !options.model) {
        options.model = property;
      }

      if (options.model) {
        installModelInterceptor(setAttributeFn, {
          target,
          property,
          descriptor,
          model: options.model
        });
      }

      ['get', 'set'].forEach(k => {
        const orig = definition[k];

        definition[k] = function (...args) {
          attachInterceptors(this);
          return orig.call(this, ...args);
        };
      });
      return definition;
    }

    function buildAttributeDecorator(type, args, setAttributeFn) {
      if ('object' === typeof args[0] && 1 === args.length) {
        return (target, property, descriptor) => addAttribute(type, setAttributeFn, target, property, descriptor, args[0]);
      }

      return addAttribute(type, setAttributeFn, ...args, {});
    }

    const hasMany = (...args) => buildAttributeDecorator('hasMany', args, onHasManySet);

    const belongsTo = (...args) => buildAttributeDecorator('belongsTo', args, onPropertySet);

    const field = (...args) => buildAttributeDecorator('field', args, onPropertySet);

    const session = (...args) => buildAttributeDecorator('session', args, onPropertySet);

    const identifier = (...args) => buildAttributeDecorator('identifier', args, onPropertySet);

    exports.registerModel = registerModel;
    exports.findModel = findModel;
    exports.rememberModelUsing = rememberModelUsing;
    exports.lookupModelUsing = lookupModelUsing;
    exports.getSchema = getSchema;
    exports.registerCustomType = registerCustomType;
    exports.identifiedBy = identifiedBy;
    exports.unresolvedAssociations = unresolvedAssociations;
    exports.field = field;
    exports.session = session;
    exports.belongsTo = belongsTo;
    exports.hasMany = hasMany;
    exports.identifier = identifier;
    exports.createCollection = createCollection;

    Object.defineProperty(exports, '__esModule', { value: true });

})));
